# core/dispatch.py
#
#

""" basic package for the program. """

__copyright__ = "Copyright 2014 B.H.J Thate"

## IMPORTS

from core.utils import get_plugname, get_name, get_strace, error
from core.errors import NoFunction
from core.object import Object

import collections
import threading
import logging
import random
import queue
import time

## DEFINES

threads = []

## RUNNER

class Runner(threading.Thread):

    """ the working unit of CORED.  """

    _nr = 0
    
    def __init__(zelf, *args, **kwargs):
        threading.Thread.__init__(zelf, None, zelf._loop, "Runner-%s" % str(zelf._nr), args, kwargs)
        zelf.setDaemon(True)
        zelf.type = "Runner"
        zelf._queue = queue.Queue()
        zelf._ready = threading.Event() 
        zelf._status = "starting"
        zelf._nr += 1

    def _loop(zelf):
        """ loop that executes the (function, object) pairs. """
        while zelf._status:
            args, kwargs = zelf._queue.get()
            func = args[0]
            if not func: raise NoFunction()
            try: arguments = args[1:]
            except IndexError: arguments = []
            name = get_name(func)
            zelf.setName(name)
            try: func(*arguments, **kwargs)
            except: error(*arguments)
            if arguments:
                event = arguments[0]
                if "_parsed" in arguments[0]: logging.info("! %s/%s %s" % (zelf.type, "parsed", str(arguments[0]._parsed)))
                if "_ready" in arguments[0]: arguments[0].ready()
        zelf.ready()
       
    def put(zelf, *args, **kwargs):
        """ put arguments/kwargs to the Runner. """
        zelf._queue.put((args, kwargs))
        return zelf

    ## WAITERS

    def ready(zelf):
        """ signal to ready state. """
        logging.debug("! %s/ready %s" % (zelf.type, get_name(zelf)))
        zelf._ready.set()

    def clear(zelf):
        """ clear the ready state. """
        logging.debug("! %s/clear %s" % (zelf.type, get_name(zelf)))
        zelf._ready.clear()

    def wait(zelf, sec=180.0):
        """ wait for ready state. """
        logging.debug("! %s/wait %s" % (zelf.type, get_name(zelf)))
        try: zelf._ready.wait(sec)
        except: pass

## DISPATCH

class Dispatcher(Object):

    """ the dispatcher delegates the workload to the Runners. Runner gets instantiated when needed. """

    def register(zelf, *args, **kwargs):
        """ callback type with corresponding callback function. """
        o = Object()
        o.cmnd = args[0]
        o.func = args[1]
        o.plugname = get_plugname()
        if o.cmnd not in zelf: zelf[o.cmnd] = []
        zelf[o.cmnd].append(o)
        logging.debug("! %s/register %s" % (zelf.type, get_name(o.func)))

    def dispatch(zelf, *args, **kwargs):
        """ dispatch an event onto a Runner. """
        from core import kernel
        cmnd = args[0]
        event = args[1]
        if cmnd not in zelf: return
        for cmnd in zelf[cmnd]:
            logging.info("! %s/dispatch(%s) %s %s" % (zelf.type, event.type, event.origin, get_name(cmnd.func)))
            thr = kernel.workers.put(cmnd.func, event) 
        return thr

    def put(zelf, *args, **kwargs):
        """ put load to the Runner. """
        if not args: logging.error("no arguments for dispatch") ; return
        target = zelf.make_new()
        return target.put(*args, **kwargs)

    def make_new(zelf, *args, **kwargs):
        """ create a Runner. """
        runner = Runner(*args, **kwargs)
        runner.start()
        return runner
